import pathlib
import shutil
from loophost import LOOPHOST_DOMAIN, TUNNEL_DOMAIN, TARGET_DIR, PYEX, HUBDIR, USERNAME
from string import Template
from subprocess import run


def register_tunnel(project, template_path, target_path):
    shutil.copy2(template_path, TARGET_DIR)

    d = {"CWD": TARGET_DIR,
         "USERNAME": USERNAME,
         "PYEX": PYEX,
         "HUBDIR": HUBDIR,
         "LOOPHOST_DOMAIN": LOOPHOST_DOMAIN,
         "TUNNEL_DOMAIN": TUNNEL_DOMAIN,
         "LOCAL_USER": pathlib.Path('localuser.txt').read_text(),
         "PROJECT": project}

    with open(template_path, "r") as f:
        src = Template(f.read())
        result = src.substitute(d)
        with open(target_path, "w+") as o:
            o.write(result)

    run(["launchctl", "unload", target_path], cwd=TARGET_DIR)
    run(["launchctl", "load", target_path], cwd=TARGET_DIR)


def unregister_tunnel(target_path):
    run(["launchctl", "unload", target_path], cwd=TARGET_DIR)
    shutil.rmtree(target_path, ignore_errors=True)



import json
import os
import sys
import pathlib
import subprocess
from subprocess import run
from fling_cli.auth import gh_authenticate
from loophost.launchd_plist import register_tunnel
from loophost import (
    LOOPHOST_DOMAIN,
    TUNNEL_DOMAIN,
    TARGET_DIR,
    HUBDIR,
    USERNAME,
    DATA_FILE_PATH,
)


def post_install_one():
    print("Installing LoopHost...")
    os.makedirs(pathlib.Path(TARGET_DIR, "certs"), exist_ok=True)
    os.chdir(TARGET_DIR)
    with open("localuser.txt", "w+") as userfile:
        userfile.write(os.getlogin())
    authenticate_with_fling()
    issue_certs()
    create_update_loophost_json()
    register_tunnel(
        "flinghub",
        pathlib.Path(HUBDIR, "plist", "loophost.plist.template"),
        pathlib.Path(
            pathlib.Path.home(), "Library/LaunchAgents/dev.fling.hub.local.plist"
        ),
    )

    restart_as_sudo()


def post_install_two():
    setup_launchd_scripts()


def authenticate_with_fling():
    global USERNAME
    gh_authenticate()
    if not os.path.exists(pathlib.Path(TARGET_DIR, "flinguser.txt")):
        raise Exception("Github authentication failed, can't continue install.")
    with open(pathlib.Path(TARGET_DIR, "flinguser.txt"), "r") as userfile:
        USERNAME = userfile.read()


def issue_certs():
    global USERNAME, TARGET_DIR
    print("Generating SSL certificates (this may take a minute)...")
    cmd = " ".join(
        [
            "certbot",
            "certonly",
            "--config-dir ./",
            "--work-dir ./",
            "--logs-dir ./",
            "--non-interactive",
            "--expand",
            "--agree-tos",
            f"-m webmaster@{LOOPHOST_DOMAIN}",
            "--authenticator=fling_authenticator",
            f'-d "*.{USERNAME}.{LOOPHOST_DOMAIN}"',
            f'-d "{USERNAME}.{LOOPHOST_DOMAIN}"',
            # "--force-renewal",
            "--fling_authenticator-propagation-seconds=15",
        ]
    )
    print(cmd)
    run(
        cmd,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        stdin=subprocess.PIPE,
        cwd=TARGET_DIR,
        shell=True,
        check=True,
    )


def create_update_loophost_json():
    data = None
    if os.path.exists(DATA_FILE_PATH):
        with open(DATA_FILE_PATH, "r") as datafile:
            data = json.loads(datafile.read())
    else:
        data = {"apps": {}, "share": {}}
    data.update(
        {
            "fqdn": f"{USERNAME}.{LOOPHOST_DOMAIN}",
            "tunnel": f"{USERNAME}.{TUNNEL_DOMAIN}",
        }
    )
    with open(DATA_FILE_PATH, "w") as datafile:
        datafile.write(json.dumps(data))
    return data


def setup_launchd_scripts():
    register_tunnel(
        "flinghub",
        pathlib.Path(HUBDIR, "plist", "hub.plist.template"),
        "/Library/LaunchDaemons/dev.fling.hub.plist",
    )
    print("All done.")


def restart_as_sudo():
    global USERNAME
    print(
        "Switching to root user to install web services (you will be prompted for your password)"
    )
    run(
        "sudo python3 -m loophost.postinstall",
        shell=True,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        stdin=subprocess.PIPE,
        cwd=TARGET_DIR,
    )
    print("All finished.")
    run(f"""/usr/bin/open "https://{USERNAME}.{LOOPHOST_DOMAIN}" """, shell=True)
    sys.exit()


if __name__ == "__main__":
    # TODO: Make this impotent and reentrant safe
    if os.geteuid() == 0:
        post_install_two()
    else:
        post_install_one()



__version__ = '0.2.3'

import os
import pathlib
import sys


LOOPHOST_DOMAIN = "loophost.dev"
TUNNEL_DOMAIN = "fling.team"

if os.geteuid() != 0:
    os.makedirs(pathlib.Path(pathlib.Path.home(), ".flingdev"), exist_ok=True)
    os.chdir(pathlib.Path(pathlib.Path.home(), ".flingdev"))

TARGET_DIR = pathlib.Path(os.path.abspath(os.path.curdir))
PYEX = sys.executable
HUBDIR = os.path.dirname(os.path.realpath(__file__))
USERNAME = None
if os.path.exists(pathlib.Path(TARGET_DIR, "flinguser.txt")):
    with open(pathlib.Path(TARGET_DIR, "flinguser.txt"), "r") as userfile:
        USERNAME = userfile.read()

DATA_FILE_PATH = pathlib.Path(TARGET_DIR, "loophost.json")



"""Loophost administration web app.
This runs as a launchd-dispatched python web app.
It displays a web-based administration screen.
Must be run with a CWD of the .flingdev folder in the user's home directory.
"""

import os
import json
import pathlib
from fling.start import start
from flask import Flask, request, render_template, redirect
from flask_bootstrap import Bootstrap5
from flask_caching import Cache
from loophost.launchd_plist import register_tunnel, unregister_tunnel
from loophost import HUBDIR, __version__, DATA_FILE_PATH
from lastversion import has_update
from fling_cli import get_fling_client
from fling_client.api.loophost import expose_app_expose_app_put


config = json.loads(pathlib.Path.read_text(DATA_FILE_PATH))


admin = Flask(__name__)
bootstrap = Bootstrap5(admin)
cache = Cache(config={"CACHE_TYPE": "SimpleCache"})
cache.init_app(admin)
admin.register_blueprint(start, url_prefix="/start")  # unused but I need the templates


@cache.cached(timeout=3600)
def latest_version():
    return has_update(repo="loophost", at="pip", current_version=__version__)


@admin.context_processor
def inject_globals():
    return dict(NEW_VERSION=latest_version())


@admin.route("/unbind")
def unbind():
    project = request.args.get("project")
    if project and project in config["apps"]:
        del config["apps"][project]
        with open(DATA_FILE_PATH, "w+") as appjson:
            appjson.write(json.dumps(config))
    return redirect("/")


@admin.route("/share")
def share():
    project = request.args.get("project")
    target = pathlib.Path(
        pathlib.Path.home(),
        "Library",
        "LaunchAgents",
        f"dev.fling.hub.ssh.{project}.plist",
    )
    if project and project in config.get("share", {}):
        del config["share"][project]
        unregister_tunnel(target)
    elif project:
        if not config.get("share"):
            config["share"] = {}
            # TODO(JMC): Don't do this so often
            fling_client = get_fling_client(require_auth=True)
            expose_app_expose_app_put.sync(client=fling_client, app_name=project)
        config["share"][project] = "public"
        register_tunnel(
            project, pathlib.Path(HUBDIR, "plist", "ssh.plist.template"), target
        )
    with open(DATA_FILE_PATH, "w+") as appjson:
        appjson.write(json.dumps(config))
    return redirect("/")


@admin.route("/", defaults={"path": ""}, methods=["GET", "POST"])
@admin.route("/<path:path>", methods=["GET", "POST"])
def admin_page(path):
    fqdn = request.host.split(".")
    if len(fqdn) < 4:
        return render_template(
            "admin.html",
            config=config,
            apps=config["apps"],
            share=config.get("share"),
            fqdn=".".join(fqdn),
        )
    project = fqdn.pop(0)
    if request.form.get("application_port"):
        config["apps"][project] = request.form.get("application_port")
        with open(DATA_FILE_PATH, "w+") as appjson:
            appjson.write(json.dumps(config))
        return redirect(f"https://{project}.{'.'.join(fqdn)}")

    return render_template(
        "local.html",
        config=config,
        apps=config["apps"],
        share=config.get("share"),
        project=project,
        fqdn=".".join(fqdn),
    )


if __name__ == "__main__":
    admin.run(host=f"unix://{os.getcwd()}/loophost.soc")
